from rest_framework import generics, permissions

from django.db.models import Q

from core.models.messages import DirectMessage

from api.serializers import (
    DirectMessageSendSerializer,
    DirectMessageSerializerReceived,
    DirectMessageSerializerReceivedShort,
    DirectMessageSerializerSent,
    DirectMessageSerializerSentShort,
)
from api.views.mixins import ConferenceSlugMixin


class DirectMessagesReceived(ConferenceSlugMixin, generics.ListAPIView):
    permission_classes = [permissions.IsAuthenticated]
    serializer_class = DirectMessageSerializerReceivedShort

    def get_queryset(self):
        return (
            DirectMessage.objects.filter(conference=self.conference, recipient=self.request.user, deleted_by_recipient=False)
            .select_related('recipient', 'sender')
            .order_by('-timestamp')
        )


class DirectMessagesSent(ConferenceSlugMixin, generics.ListAPIView):
    permission_classes = [permissions.IsAuthenticated]
    serializer_class = DirectMessageSerializerSentShort

    def get_queryset(self):
        return (
            DirectMessage.objects.filter(conference=self.conference, sender=self.request.user, deleted_by_sender=False)
            .select_related('recipient', 'sender')
            .order_by('-timestamp')
        )


class DirectMessageReceived(ConferenceSlugMixin, generics.RetrieveUpdateAPIView):
    permission_classes = [permissions.IsAuthenticated]
    serializer_class = DirectMessageSerializerReceived

    def get_queryset(self):
        return (
            DirectMessage.objects.filter(conference=self.conference, recipient=self.request.user, deleted_by_recipient=False)
            .select_related('recipient', 'sender')
            .order_by('-timestamp')
        )


class DirectMessageSent(ConferenceSlugMixin, generics.RetrieveAPIView):
    permission_classes = [permissions.IsAuthenticated]
    serializer_class = DirectMessageSerializerSent

    def get_queryset(self):
        return (
            DirectMessage.objects.filter(conference=self.conference, sender=self.request.user, deleted_by_sender=False)
            .select_related('recipient', 'sender')
            .order_by('-timestamp')
        )


class DirectMessageSend(ConferenceSlugMixin, generics.CreateAPIView):
    permission_classes = [permissions.IsAuthenticated]
    serializer_class = DirectMessageSendSerializer

    def perform_create(self, serializer: DirectMessageSendSerializer):
        serializer.save(
            conference=self.conference,
            sender=self.request.user,
            deleted_by_recipient=self.request.user != serializer.validated_data['recipient'] and self.request.user.shadow_banned,
        )


class DirectMessageDelete(ConferenceSlugMixin, generics.DestroyAPIView):
    permission_classes = [permissions.IsAuthenticated]

    def get_queryset(self):
        return DirectMessage.objects.filter(
            Q(conference=self.conference, recipient=self.request.user, deleted_by_recipient=False)
            | Q(conference=self.conference, sender=self.request.user, deleted_by_sender=False)
        )

    def perform_destroy(self, instance: DirectMessage):
        if instance.sender == self.request.user:
            if instance.deleted_by_recipient:
                instance.delete()
                return

            instance.deleted_by_sender = True
            instance.save(update_fields=['deleted_by_sender'])

        elif instance.recipient == self.request.user:
            if instance.deleted_by_sender:
                instance.delete()
                return

            instance.deleted_by_recipient = True
            instance.save(update_fields=['deleted_by_recipient'])
