from django.urls import reverse

from core.models import Assembly, AssemblyLink, AssemblyMember
from core.models.conference import ConferenceMember
from core.models.users import PlatformUser

from backoffice.tests import BackOfficeTestCase


class AssemblyListViewTest(BackOfficeTestCase):
    def setUp(self):
        super().setUp()
        self.assemblies = [Assembly(slug=a, name=a, is_virtual=True, conference_id=self.conf.id, state=Assembly.State.ACCEPTED) for a in ('a1', 'a2', 'a3')]
        for a in self.assemblies:
            a.save()
        self.assembly_links = [AssemblyLink(a=self.assemblies[0], b=self.assemblies[1]), AssemblyLink(a=self.assemblies[0], b=self.assemblies[2])]
        for al in self.assembly_links:
            al.save()

    def test_slugname_related_assemblies(self):
        self.client.force_login(self.staff)
        resp = self.client.get(reverse('backoffice:assemblieslist', kwargs={'variant': 'slugname'}))
        self.assertEqual(resp.context['data'][0][3], 'a2, a3')
        self.assertIn(b'a2, a3', resp.content)


class AssemblyCreateViewTest(BackOfficeTestCase):
    def setUp(self):
        super().setUp()

    def test_create_assembly(self):
        resp = self.client.get(reverse('backoffice:assembly-create'))
        self.assertRedirects(resp, reverse('backoffice:login') + '?next=' + reverse('backoffice:assembly-create'))

        # Test with regular user
        self.client.force_login(self.non_staff_user)
        resp = self.client.get(reverse('backoffice:assembly-create'))
        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed(resp, 'backoffice/assembly_create.html')
        resp = self.client.post(reverse('backoffice:assembly-create'), data={'slug': 'a1', 'name': 'a1', 'is_physical': 'true', 'conference': self.conf.id})
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(Assembly.objects.count(), 1)
        assembly = Assembly.objects.first()
        self.assertIsNotNone(assembly)
        assert assembly is not None
        self.assertRedirects(resp, reverse('backoffice:assembly-edit', kwargs={'pk': assembly.pk}))

        # Test with staff user
        self.client.force_login(self.staff)
        resp = self.client.get(reverse('backoffice:assembly-create'))
        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed(resp, 'backoffice/assembly_create.html')
        resp = self.client.post(reverse('backoffice:assembly-create'), data={'slug': 'a2', 'name': 'a2', 'is_physical': 'true', 'conference': self.conf.id})
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(Assembly.objects.count(), 2)
        assembly = Assembly.objects.filter(name='a2').first()
        self.assertIsNotNone(assembly)
        assert assembly is not None
        self.assertRedirects(resp, reverse('backoffice:assembly-edit', kwargs={'pk': assembly.pk}))

    def test_create_assembly_non_public_conf(self):
        self.conf.is_public = False
        self.conf.save()
        resp = self.client.get(reverse('backoffice:assembly-create'))
        self.assertRedirects(resp, reverse('backoffice:login') + '?next=' + reverse('backoffice:assembly-create'))

        # Test with regular user, as the conference is not public, the user should not be able to create an assembly
        self.client.force_login(self.non_staff_user)
        resp = self.client.get(reverse('backoffice:assembly-create'))
        self.assertContains(resp, 'Forbidden', status_code=403)

        # Test with staff user
        self.client.force_login(self.staff)
        session = self.client.session
        session['conference'] = str(self.conf.slug)
        session.save()
        resp = self.client.get(reverse('backoffice:assembly-create'))
        self.assertEqual(resp.status_code, 200)
        self.assertTemplateUsed(resp, 'backoffice/assembly_create.html')
        resp = self.client.post(reverse('backoffice:assembly-create'), data={'slug': 'a1', 'name': 'a1', 'is_physical': 'true', 'conference': self.conf.id})
        self.assertEqual(resp.status_code, 302)
        self.assertEqual(Assembly.objects.count(), 1)
        assembly = Assembly.objects.filter(name='a1').first()
        self.assertIsNotNone(assembly)
        assert assembly is not None
        self.assertRedirects(resp, reverse('backoffice:assembly-edit', kwargs={'pk': assembly.pk}))


class AssemblyMembersViewTest(BackOfficeTestCase):
    def setUp(self):
        super().setUp()
        self.assembly = Assembly.objects.create(slug='a1', name='a1', is_physical=True, conference_id=self.conf.id, state=Assembly.State.ACCEPTED)
        self.users: list[PlatformUser] = [
            PlatformUser.objects.create(
                username=a,
            )
            for a in ('a1', 'a2', 'a3')
        ]
        members: list[ConferenceMember] = []
        for user in self.users:
            members.append(ConferenceMember.objects.create(user=user, conference=self.conf))
            user.associated_assemblies.add(AssemblyMember.objects.create(member=user, assembly=self.assembly))

    def test_member_list(self):
        self.client.force_login(self.staff)
        resp = self.client.get(reverse('backoffice:assembly-members', kwargs={'pk': self.assembly.id}))
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.context['object_list'].count(), 3)
