import logging
from datetime import date
from typing import Any

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import PermissionDenied
from django.db.models import QuerySet
from django.http import Http404, HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect, render
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import get_language, gettext
from django.utils.translation import gettext_lazy as _
from django.views.generic import TemplateView, View
from django.views.generic.detail import SingleObjectMixin
from django.views.generic.edit import CreateView, UpdateView

from core.models import ActivityLogChange, ActivityLogEntry
from core.models.assemblies import Assembly, AssemblyLink
from core.models.conference import ConferenceExportCache, ConferenceMember
from core.models.invitation import Invitation
from core.models.tags import ConferenceTag

from backoffice.forms import (
    AssemblyCreateForm,
    AssemblyEditForm,
)
from backoffice.views.mixins import AssemblyMixin, AssemblyRuleMixin, ConferenceLoginRequiredMixin

logger = logging.getLogger(__name__)


class AssemblyListView(ConferenceLoginRequiredMixin, View):
    require_conference = True

    def get(self, *args, **kwargs):
        if self.request.user.is_authenticated:
            my_assemblies = list(Assembly.objects.associated_with_user(conference=self.conference, user=self.request.user, staff_can_see=False))

            # remove stored backlink for assembly pages from the session if it is set, so the backlink will go to the overview, which is the default
            self.request.session.pop('assembly_back', None)
        else:
            my_assemblies = None

        ctx = self.get_context_data()
        ctx.update(
            {
                'active_page': 'home',
                'my_assemblies': my_assemblies,
            }
        )

        return render(self.request, 'backoffice/index.html', ctx)


class AssemblyCreateView(ConferenceLoginRequiredMixin, CreateView):
    model = Assembly
    form_class = AssemblyCreateForm

    require_conference = True
    require_conference_open = True
    template_name = 'backoffice/assembly_create.html'

    def dispatch(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        if not request.user.is_authenticated:
            return redirect(f'{reverse("backoffice:login")}?next={reverse("backoffice:assembly-create")}')
        try:
            member = ConferenceMember.objects.get(conference=self.conference, user=request.user)
        except ConferenceMember.DoesNotExist:
            raise PermissionDenied
        if (member.is_staff and member.has_perms('core.assembly_team')) or self.conference.is_open:
            return super().dispatch(request, *args, **kwargs)

        raise PermissionDenied

    def get_context_data(self, *args, **kwargs):
        return {
            **super().get_context_data(*args, **kwargs),
            'registration_closed': self.require_conference_open and not self.conference.is_open and not self.is_assembly_team,
        }

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['conference'] = self.conference
        return kwargs

    def form_valid(self, form):
        form.instance.conference = self.conference
        form.instance.state = Assembly.State.PLANNED

        form.instance.ensure_single_type_conference_match()

        form.instance.save()

        # log the action
        messages.success(self.request, _('assembly__created'))
        logger.info(
            'Assembly "%(name)s" (%(id)s) created by %(user)s}.', {'name': form.instance.name, 'id': form.instance.pk, 'user': self.request.user.username}
        )
        form.instance.log_activity(
            user=self.request.user,
            kind=ActivityLogEntry.Kind.ENTITY,
            slug=ActivityLogChange(new=form.instance.slug),
            name=ActivityLogChange(new=form.instance.name),
            state=ActivityLogChange(new=form.instance.state),
        )

        # add current user as first member of the new assembly (so that editing is possible)
        form.instance.members.create(
            member=self.request.user,
            is_representative=True,
            can_manage_assembly=True,
            show_public=True,
        )
        form.instance.log_activity(
            user=None,
            kind=ActivityLogEntry.Kind.SYSTEM,
            **{
                'members.added': ActivityLogChange(
                    new=self.request.user.username,
                ),
                f'members[{self.request.user.username}].is_representative': ActivityLogChange(new='yes'),
                f'members[{self.request.user.username}].can_manage_assembly': ActivityLogChange(new='yes'),
                f'members[{self.request.user.username}].show_public': ActivityLogChange(new='yes'),
            },
        )

        # invalidate export caches where necessary
        ConferenceExportCache.signal_assembly_modification(conference=self.conference, assembly=form.instance)

        # we're done saving the initial version, redirect to edit view
        return redirect(reverse('backoffice:assembly-edit', kwargs={'pk': form.instance.id}))


class AssemblyUpdateView(AssemblyMixin, UpdateView):
    form_class = AssemblyEditForm
    require_conference = True
    assembly_management = True
    template_name = 'backoffice/assembly_edit.html'

    def get_object(self, *args, **kwargs):
        return self.assembly

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['instance'] = self.assembly
        kwargs['staff_access'] = self.staff_access
        kwargs['staff_mode'] = self.staff_mode
        kwargs['assembly_staff_access'] = self.assembly_staff_access

        return kwargs

    def form_valid(self, form):
        # backup current values, if possible
        old_values = {fld_name: form.get_initial_for_field(field, fld_name) for fld_name, field in form.fields.items()}

        # prepare changes on the assembly
        assembly = form.save(commit=False)

        # note down changed model fields
        changes = {}
        for field in form.changed_data:
            if field in ['tags']:
                continue
            new_value = getattr(assembly, field)
            changes[field] = (old_values[field], new_value) if field in old_values else new_value

        # update additional fields (if present)
        af = {}
        old_af = assembly.registration_data or {}
        for field in self.conference.additional_fields_schema or []:
            field_type = field.get('type')
            key = field.get('key')
            val = self.request.POST.get('additional_fields[' + key + ']')

            if field_type == 'boolean':
                if val is not None:
                    # value being present means checkbox has been set
                    val = True
                else:
                    # store False if value was previously set, otherwise leave it as None
                    val = False if old_af.get(key) else None

            elif field_type == 'integer':
                try:
                    if val:
                        val = int(val)

                        if (min_value := field.get('min_value')) is not None:
                            if val < min_value:
                                val = min_value
                                messages.warning(self.request, _('Assembly__edit__cf_int_below_min') + ' ' + field.get('caption', {}).get(get_language(), key))

                        if (max_value := field.get('max_value')) is not None:
                            if val > max_value:
                                val = max_value
                                messages.warning(self.request, _('Assembly__edit__cf_int_above_max') + ' ' + field.get('caption', {}).get(get_language(), key))

                    else:
                        val = None

                except ValueError:
                    val = None
                    messages.warning(self.request, _('Assembly__edit__cf_int_error') + ' ' + field.get('caption', {}).get(get_language(), key))

            elif field_type == 'date':
                try:
                    val = date.fromisoformat(val) if val else None
                except ValueError:
                    val = None
                    messages.warning(self.request, _('Assembly__edit__cf_date_error') + ' ' + field.get('caption', {}).get(get_language(), key))

                if val is not None:
                    val_daterangeerror = False
                    if (min_value := field.get('min_value')) is not None:
                        try:
                            min_value = date.fromisoformat(min_value)
                            if val < min_value:
                                val = min_value
                                val_daterangeerror = True
                        except ValueError:
                            # min_value could not be parsed, ignore that
                            pass

                    if (max_value := field.get('max_value')) is not None:
                        try:
                            max_value = date.fromisoformat(max_value)
                            if val > max_value:
                                val = max_value
                                val_daterangeerror = True
                        except ValueError:
                            # max_value could not be parsed, ignore that
                            pass

                    if val_daterangeerror:
                        messages.warning(self.request, _('Assembly__edit__cf_date_range_error') + ' ' + field.get('caption', {}).get(get_language(), key))

                val = val.isoformat() if val else None

            elif field_type == 'option':
                if len(val.strip()) == 0:
                    if self.request.POST.get('additional_fields[' + key + '_option]') is None:
                        val = None
                    else:
                        messages.warning(self.request, _('Assembly__edit__cf_option_missing_details') + ' ' + field.get('caption', {}).get(get_language(), key))

            if (old_val := old_af.get(key)) != val:
                changes[f'registration_data.{key}'] = (old_val, val)
            af[key] = val
        assembly.registration_data = af

        new_tags = []

        # update tags: go through supplied list of tags
        given_tags = form.cleaned_data.get('tags').split(',')
        for raw_tag in given_tags:
            tag = raw_tag.strip().lower()  # type: str
            if len(tag) == 0:
                # skip empty ones
                continue
            new_tags.append(tag)
        new_tags, removed_tags = assembly.update_tags(new_tags, auto_create=True)
        # log changes in tags
        if new_tags or removed_tags:
            changes['tags'] = (' '.join(f'-{tag}' for tag in removed_tags) + ' ' + ' '.join(f'+{tag}' for tag in new_tags)).strip()

        # auto-set physical/virtual/remote type if mandated by conference
        for fld, change in assembly.ensure_single_type_conference_match().items():
            changes[fld] = change

        # auto-advance assembly states from 'planned' to 'registered' on (first) save
        if assembly.state == Assembly.State.PLANNED and (assembly.user_can_manage(self.request.user) or not self.staff_mode):
            assembly.state = Assembly.State.REGISTERED
            changes['state'] = (Assembly.State.PLANNED, Assembly.State.REGISTERED)

        # log latest update timestamp
        if self.staff_mode and not assembly.user_can_manage(self.request.user):
            assembly.last_update_staff = timezone.now()
            changes['last_update_staff'] = str(assembly.last_update_staff)
        else:
            assembly.last_update_assembly = timezone.now()
            changes['last_update_assembly'] = str(assembly.last_update_assembly)

        # save the update data
        assembly.log_activity(
            user=self.request.user,
            kind=ActivityLogEntry.Kind.STAFF if self.staff_access else ActivityLogEntry.Kind.ENTITY,
            context='backoffice.assembly',
            **{k: (ActivityLogChange(old=str(v[0]), new=str(v[1])) if isinstance(v, tuple) else str(v)) for k, v in changes.items()},
        )
        assembly.save()

        # log the action
        messages.success(self.request, _('Assembly__edit__success'))
        logger.info(
            'Assembly "%(slug)s" (%(pk)s) edited by <%(user)s>: %(changes)s',
            {
                'slug': assembly.slug,
                'pk': assembly.pk,
                'user': self.request.user.username,
                'changes': changes,
            },
        )

        # invalidate export caches where necessary
        ConferenceExportCache.signal_assembly_modification(conference=self.conference, assembly=assembly)

        # we're done saving, redirect to edit view again
        return redirect('backoffice:assembly-edit', pk=assembly.id)

    def form_invalid(self, form):
        if settings.DEBUG:
            messages.warning(self.request, f'form_invalid: {form.errors}')
        return super().form_invalid(form)

    def get_context_data(self, *args, **kwargs):
        form = self.get_form()

        base_fields = [form[x] for x in AssemblyEditForm.Meta.fields]

        context = super().get_context_data(*args, **kwargs)
        context['base_fields'] = base_fields
        context['clusters'] = Assembly.objects.filter(conference=self.conference, hierarchy=Assembly.Hierarchy.CLUSTER).order_by('name').all()
        context['public_tags'] = ConferenceTag.objects.filter(conference=self.conference, is_public=True).order_by('slug').values_list('slug', flat=True)
        return context


class AssemblyChildrenUpdateView(AssemblyMixin, TemplateView):
    template_name = 'backoffice/assembly_editchildren.html'
    require_conference = True
    assembly_management = True

    def get_object(self, *args, **kwargs):
        assembly = self.assembly
        if not Assembly.type_is(assembly):  # pragma: no cover
            raise ValueError('Object is not an Assembly')
        # say 'Not Found' if assembly is not a cluster or clusters aren't supported at all
        if not assembly.is_cluster or not self.conference.support_clusters:
            raise Http404

        # bail out if the current user is not associated as a contact
        if not assembly.user_can_manage(self.request.user, staff_can_manage=True):
            raise PermissionDenied

        return assembly

    def get_context_data(self, *args, **kwargs):
        excluded = [*list(self.assembly.children.values_list('pk', flat=True)), self.assembly.pk]
        candidates_qs = Assembly.objects.associated_with_user(
            conference=self.conference, user=self.request.user, show_public=True, staff_can_see=False
        ).exclude(pk__in=excluded)
        candidates = list(candidates_qs.order_by('name'))
        return {
            **super().get_context_data(*args, **kwargs),
            'candidates': candidates,
            'received_invitations': self.assembly.received_invitations.filter(state=Invitation.RequestsState.REQUESTED, type=Invitation.InvitationType.HABITAT),
            'sent_invitations': self.assembly.sent_invitations.filter(state=Invitation.RequestsState.REQUESTED, type=Invitation.InvitationType.HABITAT),
        }


class AssemblyLinksUpdateView(AssemblyMixin, View):
    require_conference = True
    assembly_management = True

    def get_candidates(self) -> 'QuerySet[Assembly]':
        return (
            Assembly.objects.associated_with_user(conference=self.conference, user=self.request.user, show_public=True)
            .filter(hierarchy=Assembly.Hierarchy.REGULAR)
            .exclude(state__in=[Assembly.State.PLANNED, Assembly.State.HIDDEN, Assembly.State.REJECTED])
            # Remove self from list (cannot link to self)
            .exclude(pk=self.assembly.pk)
            # Remove already linked assemblies
            .exclude(related_assemblies__a=self.assembly)
        )

    def get_object(self, *args, **kwargs):
        assembly = self.assembly
        if not Assembly.type_is(assembly):  # pragma: no cover
            raise ValueError('Object is not an Assembly')

        # bail out if the current user is not associated as a contact
        if not assembly.user_can_manage(self.request.user, staff_can_manage=True):
            raise PermissionDenied

        return assembly

    def post(self, *args, **kwargs):
        request = self.request
        assembly = self.get_object()

        add_id = request.POST.get('add', None)
        remove_id = request.POST.get('delete', None)

        if add_id is not None:
            try:
                linked_assembly = self.get_candidates().get(pk=add_id)
            except Assembly.DoesNotExist:
                messages.warning(self.request, _('AssemblyLink__assembly__not_found %s') % add_id)
                linked_assembly = None

            if linked_assembly is not None:
                if linked_assembly.pk == assembly.pk:
                    messages.info(request, 'AssemblyLink__cannot_self_link')
                    return redirect('backoffice:assembly-editlinks', pk=assembly.pk)

                linked_assemblies_old = len(self.assembly.linked_assemblies)
                _link, link_created = assembly.assembly_links.get_or_create(b=linked_assembly, is_public=True, type=AssemblyLink.Type.RELATED)
                if link_created:
                    messages.success(request, gettext('Assembly__edit__link__added').format(linked_name=linked_assembly.name))
                    linked_assemblies = len(self.assembly.linked_assemblies)
                    self.assembly.log_activity(
                        user=self.request.user,
                        kind=ActivityLogEntry.Kind.STAFF if self.staff_access else ActivityLogEntry.Kind.ENTITY,
                        context='backoffice.assembly',
                        linked_assembly=ActivityLogChange(new=str(linked_assembly.name)),
                        linked_assemblies=ActivityLogChange(old=str(linked_assemblies_old), new=str(linked_assemblies)),
                    )

        if remove_id is not None:
            link = AssemblyLink.objects.filter(a=self.assembly, b_id=remove_id).first()
            if link is None:
                messages.warning(self.request, _('AssemblyLink__link__not_found %s') % remove_id)
            else:
                linked_assemblies_old = len(self.assembly.linked_assemblies)
                count, _model = link.delete()
                linked_assemblies = len(self.assembly.linked_assemblies)
                if count > 0:
                    messages.success(request, gettext('Assembly__edit__link__removed').format(linked_name=link.b.name))
                    self.assembly.log_activity(
                        user=self.request.user,
                        kind=ActivityLogEntry.Kind.STAFF if self.staff_access else ActivityLogEntry.Kind.ENTITY,
                        context='backoffice.assembly',
                        linked_assembly=ActivityLogChange(old=str(link.b.name)),
                        linked_assemblies=ActivityLogChange(old=str(linked_assemblies_old), new=str(linked_assemblies)),
                    )

        return redirect('backoffice:assembly-editlinks', pk=assembly.pk)

    def get(self, *args, **kwargs):
        candidates_qs = self.get_candidates()
        candidates = list(candidates_qs.order_by('name'))
        context = self.get_context_data()
        context['candidates'] = candidates
        return render(self.request, 'backoffice/assembly_editlinks.html', context)


class AssemblyDetailView(AssemblyMixin, TemplateView):
    require_conference = True
    template_name = 'backoffice/assembly_detail.html'

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        return {
            **super().get_context_data(**kwargs),
            'received_invitations': self.assembly.received_invitations.filter(state=Invitation.RequestsState.REQUESTED),
            'sent_invitations': self.assembly.sent_invitations.filter(state=Invitation.RequestsState.REQUESTED),
        }


class AssemblyParentLeaveView(AssemblyRuleMixin, SingleObjectMixin, View):
    model = Assembly
    template_name_suffix = '_leave'
    fields = ['id']
    permission_required = ['core.leave_habitat_assembly']

    def get_queryset(self):
        id_list = Assembly.objects.manageable_by_user(user=self.request.user, conference=self.conference).values_list('id', flat=True)
        id_list = id_list.union(Assembly.objects.filter(parent__in=id_list).values_list('id', flat=True))
        return Assembly.objects.filter(id__in=id_list)

    def post(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        assembly = self.get_object()
        if not Assembly.type_is(assembly):  # pragma: no cover
            raise ValueError('Object is not an Assembly')
        if assembly.parent is None:
            messages.error(request, _('Assembly__parent_leave__msg_no_parent'))
            return HttpResponseRedirect(reverse('backoffice:assembly', kwargs={'pk': assembly.pk}))

        old_parent = assembly.parent
        old_children = ', '.join(old_parent.children.all().values_list('name', flat=True))
        assembly.parent = None
        assembly.save()
        messages.success(request, _('Assembly__parent_leave__msg_success {name}').format(name=old_parent.name))
        assembly.log_activity(
            user=request.user,
            parent=ActivityLogChange(old=str(old_parent.name), new=str(None)),
        )
        old_parent.log_activity(
            user=request.user,
            children=ActivityLogChange(old=old_children, new=', '.join(old_parent.children.all().values_list('name', flat=True))),
        )

        if assembly.has_user(self.request.user):
            return HttpResponseRedirect(reverse('backoffice:assembly', kwargs={'pk': assembly.pk}))
        if old_parent.has_user(self.request.user):
            return HttpResponseRedirect(reverse('backoffice:assembly', kwargs={'pk': old_parent.pk}))

        # Fallback if user is not in any of the assemblies, this should not happen
        return HttpResponseRedirect(reverse('backoffice:index'))  # pragma: no cover
