import logging

from django.contrib.auth import views as auth_views
from django.http import JsonResponse
from django.urls import reverse_lazy
from django.views.generic import TemplateView, View

from core.views import BaseLoginView, BasePasswordResetConfirmView, BasePasswordResetView, BaseRegistrationActivationView, BaseRegistrationView

from .mixins import ConferenceLoginRequiredMixin, ConferenceRequiredMixin, PasswordMixin

logger = logging.getLogger(__name__)


class LoginView(BaseLoginView):
    template_name = 'backoffice/login.html'
    next_page = 'backoffice:index'


class LogoutView(auth_views.LogoutView):
    next_page = 'backoffice:index'


# TODO: Add conference name detection
class RegistrationView(BaseRegistrationView):
    success_url = reverse_lazy('backoffice:account_activation_sent')
    activation_url = 'backoffice:signup_activate'


class SignupDoneView(TemplateView):
    template_name = 'registration/signup_done.html'


class RegistrationActivationView(ConferenceRequiredMixin, BaseRegistrationActivationView):
    redirect_target_failure = 'backoffice:signup'
    redirect_target_success = 'backoffice:index'


class AuthDebugView(ConferenceLoginRequiredMixin, View):
    def get(self, *args, **kwargs):
        u = self.request.user

        result = {}
        result['conference'] = self.conference.slug
        result['global'] = {
            'userid': u.pk,
            'username': u.username,
            'active': u.is_active,
            'flags': [],
            'groups': list(u.groups.values_list('name', flat=True)),
            'teams': [f"{x['team__conference__name']}: {x['team__name']}" for x in u.teams.values('team__name', 'team__conference__name')],
            'permissions': [str(x) for x in u.get_all_permissions()],
        }
        if u.is_superuser:
            result['global']['flags'].append('superuser')
        if u.is_staff:
            result['global']['flags'].append('staff')

        if (member := self.conferencemember) is not None:
            result['member'] = {
                'staff': member.is_staff,
                'active_angel': member.active_angel,
                'pages': member.static_page_groups,
                'groups': list(member.permission_groups.values_list('name', flat=True)),
                'permissions': [str(x) for x in member.get_permission_groups_permissions()],
            }

            result['combined_permissions'] = [str(x) for x in member.all_permissions]

        else:
            result['member'] = False

        return JsonResponse(result, json_dumps_params={'indent': 2})


class PasswordChangeView(PasswordMixin, auth_views.PasswordChangeView):
    template_name = 'registration/change_password.html'
    success_url = reverse_lazy('backoffice:password_change_done')


class PasswordChangeDoneView(PasswordMixin, auth_views.PasswordChangeView):
    template_name = 'registration/change_password_done.html'


class PasswordResetView(BasePasswordResetView):
    success_url = reverse_lazy('backoffice:password_reset_done')
    extra_email_context = {'reset_url': 'backoffice:password_reset_confirm'}


class PasswordResetDoneView(PasswordMixin, auth_views.PasswordResetDoneView):
    template_name = 'registration/reset_password_done.html'


class PasswordResetConfirmView(PasswordMixin, BasePasswordResetConfirmView):
    template_name = 'registration/reset_password_confirm.html'
    success_url = reverse_lazy('backoffice:password_reset_complete')
    failure_url = reverse_lazy('backoffice:password_reset')


class PasswordResetCompleteView(PasswordMixin, auth_views.PasswordResetCompleteView):
    template_name = 'registration/reset_password_complete.html'
