import json
import logging
from datetime import datetime

from django.contrib import messages
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.models import OuterRef
from django.db.models.expressions import F
from django.db.models.functions import JSONObject
from django.shortcuts import redirect
from django.urls import reverse, reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView, TemplateView
from django.views.generic.edit import DeleteView, UpdateView

from core.models import Conference, Lock, StaticPage, StaticPageNamespace, StaticPageRevision

from backoffice.forms import StaticPageEditForm
from backoffice.views.mixins import ConferenceLoginRequiredMixin, guess_active_sidebar_item


class WikiAdminMixin(ConferenceLoginRequiredMixin):
    require_conference = True
    permission_required = ['core.static_pages']

    active_page = 'wiki'

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx['active_page'] = self.active_page

        namespace_count = StaticPageNamespace.objects.filter(conference=self.conference).count()
        page_count = StaticPage.objects.filter(conference=self.conference).count()
        ctx['sidebar'] = {
            'title': _('nav_wiki'),
            'title_link': reverse('backoffice:wiki'),
            'items': [
                {
                    'caption': _('StaticPageNamespaces'),
                    'count': namespace_count,
                    'link': reverse('backoffice:wiki-namespaces'),
                },
                {
                    'caption': _('Locks'),
                    'count': LockListView.get_pagelock_qs_for_conference(self.conference).count(),
                    'link': reverse('backoffice:wiki-locks'),
                },
                {
                    'caption': _('StaticPages'),
                    'count': page_count,
                    'link': reverse('backoffice:wiki-pages'),
                },
            ],
        }

        # try to guess 'active' sidebar item
        guess_active_sidebar_item(self.request, ctx['sidebar']['items'], with_query_string=False)

        return ctx


class WikiOverviewView(WikiAdminMixin, TemplateView):
    template_name = 'backoffice/wiki_overview.html'

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        ctx['last_edits'] = StaticPageRevision.objects.filter(page__conference=self.conference).prefetch_related('page', 'author').order_by('-timestamp')
        return ctx


class NamespaceListView(WikiAdminMixin, ListView):
    template_name = 'backoffice/wiki_namespace_list.html'

    def get_queryset(self):
        return StaticPageNamespace.objects.filter(conference=self.conference)


class LockListView(WikiAdminMixin, ListView):
    template_name = 'backoffice/wiki_lock_list.html'

    @staticmethod
    def get_pagelock_qs_for_conference(conference: Conference):
        return Lock.objects.filter(
            content_type=ContentType.objects.get_for_model(StaticPage),
            object_id__in=conference.pages.values_list('id'),
        )

    def get_queryset(self):
        return self.get_pagelock_qs_for_conference(self.conference)

    def post(self, request, *args, **kwargs):
        lock_id = request.POST.get('remove')
        lock = self.get_queryset().filter(pk=lock_id).first()
        if lock is None:
            messages.info(request, _('Lock-gone'))
        else:
            lock.delete()

            locked_pageslug = lock.content_object.slug
            locked_by = lock.lock_holder.username

            messages.success(request, _('Lock-dropped') + f' (page={locked_pageslug}; by={locked_by})')
            logging.warning(
                'Lock %s on wiki page %s ("%s") for user "%s" removed by "%s".', lock_id, lock.object_id, locked_pageslug, locked_by, request.user.username
            )

        return redirect('backoffice:wiki-locks')


class LastRevisionDetailsDecoder(json.JSONDecoder):
    def decode(self, s: str, *args, **kwargs):
        res = super().decode(s, *args, **kwargs)

        # parse timestamp. Will parse microseconds wrongly (psql returns a fraction of a second, not the total microseconds, so for example .6 is 600000 ms)
        # -> they'll be incorrect, but we don't need this permission so we will cut them off.
        ts = datetime.strptime(res['timestamp'], r'%Y-%m-%dT%H:%M:%S.%f%z')
        ts = ts.replace(microsecond=0)

        res['timestamp'] = ts
        return res


class LastRevisionJSONObject(JSONObject):
    output_field = models.JSONField(decoder=LastRevisionDetailsDecoder)


class PagesView(WikiAdminMixin, ListView):
    template_name = 'backoffice/wiki_page_list.html'

    def get_queryset(self, *args, **kwargs):
        last_revision_details = StaticPageRevision.objects.filter(page=OuterRef('pk'), pk=OuterRef('public_revision')).values(
            data=LastRevisionJSONObject(author=F('author__username'), timestamp=F('timestamp'))
        )[0:1]
        return (
            StaticPage.objects.accessible_by_user(conference=self.conference, user=self.request.user, language=None)
            .order_by('slug', 'language')
            .annotate(last_revision_details=last_revision_details)
        )


class PageView(WikiAdminMixin, UpdateView):
    form_class = StaticPageEditForm
    template_name = 'backoffice/wiki_page_detail.html'

    def get_object(self):
        return StaticPage.objects.accessible_by_user(self.request.user, self.conference, language=None).get(pk=self.kwargs.get('pk'))

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['object'] = self.object
        context['revisions'] = self.object.revisions.filter(is_draft=False).order_by('-timestamp', '-pk')
        return context

    def form_valid(self, form):
        messages.success(self.request, _('StaticPage__metadata_changed') + ': ' + self.object.slug)
        return super().form_valid(form)

    def form_invalid(self, form):
        messages.error(self.request, _('StaticPage__metadata_not_changed') + ': ' + self.object.slug)
        return super().form_invalid(form)

    def get_success_url(self, *args, **kwargs):
        return reverse('backoffice:wiki-page-detail', kwargs={'pk': self.object.pk})


class PageDeleteView(WikiAdminMixin, DeleteView):
    model = StaticPage
    success_url = reverse_lazy('backoffice:wiki-pages')

    def get_object(self):
        return StaticPage.objects.accessible_by_user(self.request.user, self.conference, language=None).get(pk=self.kwargs.get('pk'))

    def form_valid(self, form):
        res = super().form_valid(form)
        messages.success(self.request, _('StaticPage__deleted'))
        return res


class PageRevisionDeleteView(WikiAdminMixin, DeleteView):
    model = StaticPage
    success_url = reverse_lazy('backoffice:wiki-pages')

    def get_object(self):
        page = StaticPage.objects.accessible_by_user(self.request.user, self.conference, language=None).get(pk=self.kwargs.get('pk'))
        return page.revisions.filter(revision=self.request.POST.get('revision'))

    def get_success_url(self) -> str:
        return reverse('backoffice:wiki-page-detail', kwargs={'pk': self.kwargs.get('pk')})

    def form_valid(self, form):
        res = super().form_valid(form)
        messages.success(self.request, _('StaticPageRevision__deleted'))
        return res
