import logging
from typing import Any

from django.core.exceptions import ValidationError
from django.forms import DateTimeInput
from django.utils.translation import gettext_lazy as _

from core.base_forms import TranslatedFieldsForm
from core.models.conference import Conference

logger = logging.getLogger(__name__)


class ConferencePublicationForm(TranslatedFieldsForm):
    class Meta:
        model = Conference
        fields = [
            'name',
            'slug',
            'start',
            'end',
            'publication_date',
            'timezone',
            'global_notification',
            'is_public',
        ]

        widgets = {
            'start': DateTimeInput(attrs={'type': 'datetime-local'}, format='%Y-%m-%dT%H:%M:%S%z'),
            'end': DateTimeInput(attrs={'type': 'datetime-local'}, format='%Y-%m-%dT%H:%M:%S%z'),
            'publication_date': DateTimeInput(attrs={'type': 'datetime-local'}, format='%Y-%m-%dT%H:%M:%S%z'),
        }

    def __init__(
        self,
        *args,
        instance: Conference | None = None,
        publish: bool = False,
        **kwargs,
    ):
        self.publish = publish
        self.create = instance is None
        super().__init__(*args, instance=instance, **kwargs)
        if not self.create:
            self.fields['name'].disabled = True
            self.fields['slug'].disabled = True

    def clean(self) -> dict[str, Any]:
        if self.instance.is_public != self.cleaned_data['is_public']:
            raise ValidationError(_('Conference__is_public__unchangeable'))
        self.cleaned_data['is_public'] = not self.instance.is_public if self.publish else self.instance.is_public
        return super().clean()


class ConferenceRegistrationForm(TranslatedFieldsForm):
    class Meta:
        model = Conference
        fields = [
            'name',
            'slug',
            'publication_date',
            'registration_start',
            'registration_deadline',
            'additional_fields_schema',
        ]

        widgets = {
            'publication_date': DateTimeInput(attrs={'type': 'datetime-local'}, format='%Y-%m-%dT%H:%M:%S%z'),
            'registration_start': DateTimeInput(attrs={'type': 'datetime-local'}, format='%Y-%m-%dT%H:%M:%S%z'),
            'registration_deadline': DateTimeInput(attrs={'type': 'datetime-local'}, format='%Y-%m-%dT%H:%M:%S%z'),
        }

    def __init__(
        self,
        *args,
        instance: Conference | None = None,
        **kwargs,
    ):
        super().__init__(*args, instance=instance, **kwargs)
        self.fields['name'].disabled = True
        self.fields['slug'].disabled = True
