import logging
import string
from hashlib import sha1
from random import SystemRandom
from urllib.parse import quote, urlencode, urljoin
from uuid import uuid4
from xml.etree import ElementTree as ET

import requests

from django.utils.translation import gettext as _

from core.models import BackendMixin, Event, PlatformUser, Room
from core.models.assemblies import Assembly, AssemblyMember

from .error import IntegrationError

logger = logging.getLogger(__name__)
PASSWORD_CHARS = string.ascii_letters + string.digits


def _params_to_str(data: dict[str, str | int | bool]):
    res = {}
    for k, v in data.items():
        if isinstance(v, bool):
            v = 'true' if v else 'false'  # noqa: PLW2901
        elif isinstance(v, int):
            v = str(v)  # noqa: PLW2901
        else:
            assert isinstance(v, str), f'{v!r} is no string!'
        res[k] = v
    return res


class BigBlueButtonIntegration:
    """
    This class talks with a BigBlueButton server's API.

    For detailled documentation of the API please see
    https://docs.bigbluebutton.org/dev/api.html
    """

    def __init__(self, api_url, api_token, end_meeting_callback, initial_presentation_url):
        self._api_url = api_url
        self._api_token = api_token
        self._end_meeting_callback = end_meeting_callback
        self._initial_presentation_url = initial_presentation_url
        self._session = requests.session()

    def _send_request(self, resource: str, params: dict[str, str] | None = None, raw=False, post_body=None):
        params = params or {}
        encoded_params = urlencode(_params_to_str(params))
        hash_input = resource + encoded_params + self._api_token
        hash_input = hash_input.encode('utf-8')
        checksum = sha1(hash_input).hexdigest()

        if encoded_params:
            params_str = encoded_params + '&checksum=' + checksum
        else:
            params_str = 'checksum=' + checksum
        request_url = urljoin(self._api_url, resource)

        if post_body:
            resp = self._session.post(request_url, params=params_str, allow_redirects=False, data=post_body)
        else:
            resp = self._session.get(request_url, params=params_str, allow_redirects=False)
        if raw:
            return resp

        if resp.status_code != 200:
            logger.warning('Request for resource %r failed with status code %r', resource, resp.status_code)
            raise IntegrationError(_('Request failed'))

        try:
            resp = ET.fromstring(resp.text)
        except UnicodeDecodeError:
            logger.exception('Response decoding failed')
            raise IntegrationError(_('Invalid Response'))

        return resp

    def is_available(self):
        return bool(self._api_token)

    def can_create_for_assembly(self, assembly: Assembly):
        assert assembly is not None

        # no restrictions
        return True

    def create_room(self, room: BackendMixin):
        assert room is not None and ((isinstance(room, Room) and room.room_type == Room.RoomType.BIGBLUEBUTTON) or isinstance(room, Event))

        if room.backend_status in {Room.BackendStatus.ACTIVE, Room.BackendStatus.FULL}:
            # room was already created, don't need to create it twice
            return

        if not self.is_available():
            raise IntegrationError('Currently not available.')

        room.backend_status = Room.BackendStatus.SETUP
        room.save(update_fields=['backend_status'])

        room_id = str(uuid4())
        join_pw = ''.join(SystemRandom().choices(PASSWORD_CHARS, k=32))
        mod_pw = ''.join(SystemRandom().choices(PASSWORD_CHARS, k=32))
        close_secret = ''.join(SystemRandom().choices(PASSWORD_CHARS, k=32))

        if self._end_meeting_callback:
            end_meeting_callback = f'{self._end_meeting_callback}?close_secret={quote(close_secret)}'
        else:
            end_meeting_callback = ''

        params = {
            'name': room.name,
            'meetingID': room_id,
            'attendeePW': join_pw,
            'moderatorPW': mod_pw,
            'record': False,
            'muteOnStart': False,
            'allowModsToUnmuteUsers': False,
            'meta_endCallbackUrl': end_meeting_callback,
        }

        # set BBB's max_participants based on room's capacity
        if room.capacity is not None and room.capacity > 0:
            params['maxParticipants'] = room.capacity

        request_body = None
        if self._initial_presentation_url:
            request_body = (
                f'<?xml version="1.0" encoding="UTF-8"?>'
                f'<modules>'
                f'    <module name="presentation">'
                f'        <document url="{self._initial_presentation_url}" />'
                f'    </module>'
                f'</modules>'
            )

        try:
            result = self._send_request('create', params, post_body=request_body)
        except Exception:
            room.backend_status = Room.BackendStatus.ERROR
            room.save(update_fields=['backend_status'])
            raise

        retcode = result.find('returncode')
        if retcode is not None and retcode.text != 'SUCCESS':
            room.backend_status = Room.BackendStatus.ERROR
            room.save(update_fields=['backend_status'])
            message = result.find('message')
            if message is not None:
                message = message.text
            logger.warning('Request to create Room %s failed. Message: %r', room, message)
            raise IntegrationError(_('Request failed'))

        room.backend_link = result.find('meetingID').text
        room.backend_status = Room.BackendStatus.ACTIVE
        room.backend_data = {
            'attendeePW': result.find('attendeePW').text,
            'moderatorPW': result.find('moderatorPW').text,
            'createTime': result.find('createTime').text,
            'close_secret': close_secret,
        }
        room.occupants = 0

        if room.backend_link != room_id:
            logger.warning('Requested meetingID %s but got %s', room_id, room.backend_link)
        if room.backend_data['attendeePW'] != join_pw:
            logger.warning('Requested attendeePW %s but got %s', join_pw, room.backend_data['attendeePW'])
        if room.backend_data['moderatorPW'] != mod_pw:
            logger.warning('Requested moderatorPW %s but got %s', mod_pw, room.backend_data['moderatorPW'])

        room.save(update_fields=['backend_data', 'backend_status', 'backend_link', 'occupants'])
        return result

    def remove_room(self, room: BackendMixin):
        assert (room is not None and (isinstance(room, Room) and room.room_type == Room.RoomType.BIGBLUEBUTTON)) or isinstance(room, Event)

        if room.backend_status not in {Room.BackendStatus.ACTIVE, Room.BackendStatus.FULL, Room.BackendStatus.INACTIVE}:
            return

        mod_pw = room.backend_data.get('moderatorPW')
        params = {
            'meetingID': room.backend_link,
            'password': mod_pw,
        }
        result = self._send_request('end', params)

        retcode = result.find('returncode')
        if retcode is not None and retcode.text != 'SUCCESS':
            room.backend_status = Room.BackendStatus.ERROR
            room.save(update_fields=['backend_status'])
            message = result.find('message')
            if message is not None:
                message = message.text
            logger.warning('Request to delete room %s failed. Message: %r', room, message)
            raise IntegrationError(_('Request failed'))

        room.backend_status = Room.BackendStatus.INACTIVE
        room.save(update_fields=['backend_status'])
        return result

    def _room_status(self, room: BackendMixin, commit=True):
        assert (room is not None and (isinstance(room, Room) and room.room_type == Room.RoomType.BIGBLUEBUTTON)) or isinstance(room, Event)

        if room.backend_status not in {Room.BackendStatus.ACTIVE, Room.BackendStatus.FULL, Room.BackendStatus.INACTIVE}:
            return False

        resp = self._send_request('getMeetingInfo', {'meetingID': room.backend_link})
        retcode = resp.find('returncode')
        if retcode is not None and retcode.text != 'SUCCESS':
            message_key = resp.find('messageKey')
            if message_key is not None and (message_key.text in ('notFound', 'invalidMeetingIdentifier')):
                room.backend_status = Room.BackendStatus.INACTIVE
            else:
                logger.warning('getMeetingInfo request for room %r failed with message %r', room, None if message_key is None else message_key.text)
                raise IntegrationError(_('Request failed'))

        try:
            room.occupants = int(resp.find('participantCount').text)
        except (ValueError, AttributeError):
            room.occupants = None

        if commit:
            room.save(update_fields=['backend_status', 'occupants'])

    def room_status(self, room: BackendMixin):
        return self._room_status(room, True)

    def _join_room(self, room: BackendMixin, user: PlatformUser, anonymous: bool = False, retrying: bool = False):
        assert (room is not None and (isinstance(room, Room) and room.room_type == Room.RoomType.BIGBLUEBUTTON)) or isinstance(room, Event)
        assert user is not None

        create_room = False
        if room.backend_status in {Room.BackendStatus.NEW, Room.BackendStatus.ERROR, Room.BackendStatus.NONE}:
            create_room = True
        else:
            self._room_status(room, False)
            if room.backend_status == Room.BackendStatus.FULL:
                room.save(update_fields=['backend_status', 'occupants'])
                return None
            if room.backend_status == Room.BackendStatus.INACTIVE:
                create_room = True

        if create_room:
            self.create_room(room)

        if room.backend_status != Room.BackendStatus.ACTIVE:
            return None  # can't join a nonexistent room

        conf = room.conference
        is_moderator = conf.users.filter(user=user, is_staff=True).exists()

        if not is_moderator:
            if isinstance(room, Room) and room.assembly_id:
                if AssemblyMember.objects.filter(assembly_id=room.assembly_id, member=user).exists():
                    is_moderator = True
            elif isinstance(room, Event) and room.owner_id == user.pk:
                is_moderator = True

        if is_moderator:
            join_pw = room.backend_data.get('moderatorPW')
        else:
            join_pw = room.backend_data.get('attendeePW')

        params = {
            'fullName': user.username,
            'meetingID': room.backend_link,
            'password': join_pw,
            'userID': str(user.id),
            'createTime': room.backend_data.get('createTime'),
            'redirect': True,
        }
        result = self._send_request('join', params, raw=True)
        return result.headers['Location']

    def join_room(self, room: BackendMixin, user: PlatformUser, anonymous: bool = False):
        return self._join_room(room, user, anonymous, False)
