from django.core.management.base import BaseCommand, CommandError
from django.db import transaction

from core.models import Event, Room, RoomLink, ScheduleSourceMapping


class Command(BaseCommand):
    help = "Merges two given rooms by moving everything from the 'bad' one to the 'good' one."

    def add_arguments(self, parser):
        parser.add_argument('bad_room_id', help='id of room which shall be nuked')
        parser.add_argument('good_room_id', help='id of room which shall have everything afterwards')
        parser.add_argument('--dry-run', '-n', action='store_true')

    def handle(self, *args, **options):
        bad_room_id = options.get('bad_room_id')
        good_room_id = options.get('good_room_id')
        dry_run = options.get('dry_run', False)

        # lookup bad room
        try:
            bad_room = Room.objects.get(pk=bad_room_id)
            print('BAD: ', bad_room)
        except Room.DoesNotExist:
            raise CommandError(f'The bad room with id "{bad_room_id}" could not be found.')

        # lookup good room
        try:
            good_room = Room.objects.get(pk=good_room_id)
            print('GOOD: ', good_room)
        except Room.DoesNotExist:
            raise CommandError(f'The good room with id "{good_room_id}" could not be found.')

        # sanity check
        if bad_room.conference_id != good_room.conference_id:
            raise CommandError('Both rooms must be part of the same conferene!')

        # do the actual merge
        with transaction.atomic():
            print('move links from bad to good room:', end=' ', flush=True)
            bad_room_links = RoomLink.objects.filter(room=bad_room)
            print(bad_room_links.update(room=good_room))

            print('move events from bad to good room:', end=' ', flush=True)
            bad_room_events = Event.objects.filter(room=bad_room)
            print(bad_room_events.update(room=good_room))

            print('update schedule import mappings:', end=' ', flush=True)
            bad_room_mappings = ScheduleSourceMapping.objects.filter(mapping_type=ScheduleSourceMapping.MappingType.ROOM, local_id=bad_room.pk)
            print(bad_room_mappings.update(local_id=good_room.pk))

            print('delete bad room:', end=' ', flush=True)
            print(bad_room.delete()[0])

            if dry_run:
                raise CommandError('Aborting because of --dry-run.')
