import logging
import random
from datetime import timedelta
from pathlib import Path
from re import compile as re_compile
from typing import Any
from uuid import UUID, uuid4

from ordered_set import OrderedSet

from django.conf import settings
from django.contrib.contenttypes.fields import GenericRelation
from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.db import models
from django.db.models import Q, QuerySet
from django.forms import DurationField
from django.utils import timezone
from django.utils.dateparse import parse_datetime
from django.utils.functional import cached_property
from django.utils.text import slugify
from django.utils.timezone import is_aware, make_aware
from django.utils.translation import gettext_lazy as _

from core.fields import ConferenceReference
from core.markdown import compile_translated_markdown_fields, store_relationships
from core.models.activitylog import ActivityLogMixin
from core.models.assemblies import Assembly
from core.models.base_managers import ConferenceManagerMixin
from core.models.conference import Conference, ConferenceMember, ConferenceTrack
from core.models.rooms import Room
from core.models.shared import BackendMixin
from core.models.tags import TaggedItemMixin, TagItem
from core.models.users import PlatformUser
from core.utils import download_from_url, str2bool, str2timedelta
from core.validators import ImageDimensionValidator

SIMPLE_TIME_RE = re_compile(r'(.*\s)?(\d+:\d+)$')

logger = logging.getLogger(__name__)


def get_event_banner_filename(instance: 'Event', filename: Path | str):
    ext = str(filename).rsplit('.', maxsplit=1)[-1]
    return Path('events').joinpath(f'{instance.id}.{ext}')


class EventDurationFormField(DurationField):
    def to_python(self, value):
        try:
            value = int(value)
            value = timedelta(minutes=value)
        except ValueError:
            if SIMPLE_TIME_RE.match(value):
                value = f'{value}:00'
        return super().to_python(value)

    def prepare_value(self, value):
        if isinstance(value, timedelta):
            return int(value.total_seconds() / 60)
        return value


class EventDurationField(models.DurationField):
    def formfield(self, **kwargs: Any) -> Any:
        defaults = {'form_class': EventDurationFormField}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class EventManager(ConferenceManagerMixin['Event']):
    staff_permissions = ['core.assembly_team', 'core.scheduleadmin']
    conference_filter = 'assembly__conference'
    assembly_filter = 'assembly'

    def apply_public_filter(self, queryset: 'QuerySet[Event]', member: ConferenceMember | None = None) -> 'QuerySet[Event]':
        return queryset.filter((Q(assembly__state__in=Assembly.PUBLIC_STATES) | Q(kind=Event.Kind.SELF_ORGANIZED)) & Q(is_public=True))

    def apply_assembly_rights_filter(self, queryset: QuerySet['Event'], member: ConferenceMember, *, only_manageable: bool | None) -> QuerySet['Event']:
        return super().apply_assembly_rights_filter(queryset, member, only_manageable=only_manageable)

    def apply_self_organized_rights_filter(
        self,
        queryset: QuerySet['Event'],
        member: ConferenceMember,
        *,
        only_manageable: bool | None = None,
    ) -> QuerySet['Event']:
        return queryset.filter(owner=member.user, kind=Event.Kind.SELF_ORGANIZED)


class Event(TaggedItemMixin, BackendMixin, ActivityLogMixin, models.Model):
    class Meta:
        constraints = [
            models.constraints.UniqueConstraint(fields=['conference', 'slug'], name='unique_event_slug'),
            models.constraints.CheckConstraint(condition=Q(kind='sos') | Q(assembly__isnull=False), name='event_has_assembly_or_is_sos'),
        ]
        verbose_name = _('Event')
        verbose_name_plural = _('Events')

    class Kind(models.TextChoices):
        OFFICIAL = 'official', _('Event__kind-official')  # offizielles Event
        ASSEMBLY = 'assembly', _('Event__kind-assembly')  # von einer Assembly kuratiert
        SELF_ORGANIZED = 'sos', _('Event__kind-sos')  # hat Einzelperson angelegt
        LIGHTNING = 'lightning', _('Event__kind-lightning')  # Lightning Talk

    class Recording(models.TextChoices):
        YES = 'yes', _('Event__recording-yes')
        NO = 'no', _('Event__recording-no')
        UNKNOWN = 'unknown', _('Event__recording-unknown')

    class Streaming(models.TextChoices):
        YES = 'yes', _('Event__streaming-yes')
        NO = 'no', _('Event__streaming-no')
        UNKNOWN = 'unknown', _('Event__streaming-unknown')

    objects = EventManager()

    id = models.UUIDField(default=uuid4, primary_key=True, editable=False)
    conference = ConferenceReference(related_name='events')
    # Unique together with conference
    slug = models.SlugField(max_length=150, help_text=_('Event__slug__help'), verbose_name=_('Event__slug'))
    kind = models.CharField(max_length=10, choices=Kind.choices, default=Kind.ASSEMBLY, help_text=_('Event__kind__help'), verbose_name=_('Event__kind'))

    name = models.CharField(max_length=200)
    track = models.ForeignKey(ConferenceTrack, blank=True, null=True, on_delete=models.PROTECT)
    assembly = models.ForeignKey(Assembly, related_name='events', on_delete=models.PROTECT, blank=True, null=True)
    room = models.ForeignKey(Room, blank=True, null=True, related_name='events', on_delete=models.PROTECT)
    location = models.CharField(max_length=200, blank=True, null=True, help_text=_('Event__location__help'), verbose_name=_('Event__location'))

    language = models.CharField(max_length=50, blank=True, null=True, help_text=_('Event__language__help'), verbose_name=_('Event__language'))
    abstract = models.TextField(blank=True, help_text=_('Event__abstract__help'), verbose_name=_('Event__abstract'))
    description = models.TextField(blank=True, help_text=_('Event__description__help'), verbose_name=_('Event__description'))
    description_html = models.TextField(blank=True)
    banner_image_height = models.PositiveIntegerField(blank=True, null=True)
    banner_image_width = models.PositiveIntegerField(blank=True, null=True)
    banner_image = models.ImageField(
        blank=True,
        null=True,
        upload_to=get_event_banner_filename,
        height_field='banner_image_height',
        width_field='banner_image_width',
        help_text=_('Event__banner_image__help'),
        verbose_name=_('Event__banner_image'),
        validators=[
            ImageDimensionValidator(
                min_size=(
                    settings.EVENT_BANNER_WIDTH_MINIMUM,
                    settings.EVENT_BANNER_HEIGHT_MINIMUM,
                ),
                max_size=(
                    settings.EVENT_BANNER_WIDTH_MAXIMUM,
                    settings.EVENT_BANNER_HEIGHT_MAXIMUM,
                ),
                square=False,
            ),
        ],
    )
    banner_image_url = models.URLField(blank=True, null=True, help_text=_('Event__banner_image_url__help'), verbose_name=_('Event__banner_image_url'))

    is_public = models.BooleanField(default=False, help_text=_('Event__is_public__help'), verbose_name=_('Event__is_public'))
    recording = models.CharField(
        max_length=25, choices=Recording.choices, default=Recording.UNKNOWN, help_text=_('Event__recording__help'), verbose_name=_('Event__recording')
    )
    streaming = models.CharField(
        max_length=25, choices=Streaming.choices, default=Streaming.UNKNOWN, help_text=_('Event__streaming__help'), verbose_name=_('Event__streaming')
    )
    schedule_start = models.DateTimeField(blank=True, null=True, help_text=_('Event__schedule_start__help'), verbose_name=_('Event__schedule_start'))
    schedule_duration = EventDurationField(blank=True, null=True, help_text=_('Event__schedule_duration__help'), verbose_name=_('Event__schedule_duration'))
    schedule_end = models.DateTimeField(blank=True, null=True, help_text=_('Event__schedule_end__help'), verbose_name=_('Event__schedule_end'))

    additional_data = models.JSONField(blank=True, default=dict)
    favorite_of = models.ManyToManyField(
        PlatformUser, related_name='favorite_events', help_text=_('Event__favorite_of__help'), verbose_name=_('Event__favorite_of')
    )

    tags = GenericRelation(TagItem, object_id_field='target_id', content_type_field='target_type', related_query_name='events')

    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        related_name='owned_events',
        on_delete=models.PROTECT,
        help_text=_('Event__owner__help'),
        verbose_name=_('Event__owner'),
    )
    """
    Ersteller/Eigentümer des Events, dies ist insb. für Self-Organized-Sessions relevant.
    """

    last_update = models.DateTimeField(auto_now=True)

    # if _can_manage is set, user can edit event. This is a cache variable for the can_manage property
    _can_manage = None

    def get_absolute_url(self):
        from core.templatetags.hub_absolute import hub_absolute  # pylint: disable=import-outside-toplevel

        return hub_absolute('plainui:event', event_slug=self.slug, i18n=settings.ARCHIVE_MODE)

    def can_manage(self, user=None):
        if self._can_manage is None:
            if not user or not user.is_authenticated:
                # guests are not allowed to manage anything
                self._can_manage = False
                return False

            if self.is_sos and user and self.owner == user:
                # the owner of a self-organized session can always manage it
                self._can_manage = True
                return True

            if self.assembly:
                self._can_manage = self.assembly.user_can_manage(user, staff_can_manage=True)

        return self._can_manage

    @cached_property
    def public_speakers(self) -> list:
        """Returns a list of all public speakers of this event."""
        persons = [participant.participant for participant in self.participants.filter(is_public=True, role=EventParticipant.Role.SPEAKER)]

        if len(persons) == 0 and self.kind == Event.Kind.SELF_ORGANIZED and self.owner:
            persons.append(self.owner)

        return persons

    @cached_property
    def is_imported(self) -> bool:
        """
        Checks if an import mapping exists for this Event.
        """
        from .schedules import ScheduleSourceMapping  # late import to circumvent circular dependency

        qs = ScheduleSourceMapping.objects.filter(
            schedule_source__conference=self.conference,
            mapping_type=ScheduleSourceMapping.MappingType.EVENT,
            local_id=self.pk,
        ).exclude(skip=True)
        return qs.exists()

    @cached_property
    def is_live_streamed(self) -> bool | None:
        """
        Checks if event is streamed
        """
        match self.streaming:
            case self.Streaming.YES:
                return True
            case self.Streaming.NO:
                return False
            case self.Streaming.UNKNOWN:
                return self.is_recorded
            case _:
                raise NotImplementedError(f'Unknown streaming state {self.streaming}')

    @cached_property
    def is_recorded(self) -> bool | None:
        """
        Checks if event is recorded
        """
        if self.room is None:
            return None
        if self.room.recording_state == Room.RecordingState.RECORDING_FORBIDDEN:
            return False
        match (self.recording, self.room.recording_state):
            case (self.Recording.YES, _):
                return True
            case (self.Recording.NO, _):
                return False
            case (self.Recording.UNKNOWN, Room.RecordingState.UNKNOWN):
                return None
            case (self.Recording.UNKNOWN, Room.RecordingState.RECORD_BY_DEFAULT):
                return True
            case (self.Recording.UNKNOWN, _):
                return False
            case _:
                raise NotImplementedError(f'Unknown recording state {self.recording}')

    @cached_property
    def is_sos(self) -> bool:
        """
        Checks if event is a Self Organized session
        """
        return self.kind == self.Kind.SELF_ORGANIZED

    @cached_property
    def conference_day(self) -> int:
        return self.conference.get_conference_day(self.schedule_start)

    def get_all_speaker_names(self):
        # plainui will prefetch speaker names in self.speakers => only do db lookup if we don't have that
        try:
            names = OrderedSet([speaker.participant.display_name for speaker in self.speakers])
        except AttributeError:
            names = OrderedSet()
            names |= OrderedSet(
                self.participants.filter(is_public=True, role=EventParticipant.Role.SPEAKER)
                .order_by('order', 'participant__display_name')
                .values_list('participant__display_name', flat=True)
            )

        # if the "correct" participant list is empty, check if we can derive some speaker names
        if len(names) == 0:
            # imported events might have a "persons" list in the additional data
            if self.additional_data is not None:
                for p in self.additional_data.get('persons', []):
                    name = p.get('public_name', p.get('name'))
                    if name:
                        names.add(name)

            # for SoS, use the owner as a speaker
            if self.kind == Event.Kind.SELF_ORGANIZED and self.owner:
                names.add(self.owner.username)

        return names

    def get_url(self, domain=None):
        if domain == 'media.ccc.de':
            return f'https://media.ccc.de/v/{self.id}'

        if self.additional_data is not None:
            return self.additional_data.get('url')

        return None

    @property
    def starts_in(self):
        """
        Returns a timedelta between now and the scheduled start is in the future.
        Returns false otherwise
        """
        if self.schedule_start is not None:
            now = timezone.now()
            if now > self.schedule_start:
                return False
            return self.schedule_start - now
        return False

    def __str__(self):
        return self.name

    @classmethod
    def from_dict(
        cls,
        data: dict,
        conference: Conference,
        assembly: Assembly = None,
        existing=None,
        pop_used_keys: bool = False,
        allow_kind: bool = False,
        allow_track: bool = False,
        room_lookup=None,
        speaker_lookup=None,
    ):
        """
        Loads an Event instance from the given dictionary.
        An existing event can be provided which's data is overwritten (in parts).

        :param data: a dictionary containing Event attributes' names as key (and their values)
        :param assembly: an Assembly instance to which this Event will belong
        :param existing: an instance of Event (or None to get a new one)
        :type existing: Event
        :param pop_used_keys: Remove 'used' keys from the provided data. This can be used to detect additional/errornous fields.
        :param allow_kind: Flag indicating whether the 'kind' attribute may be set (e.g. to 'official').
        :param allow_track: Flag indicating whether the 'track' attribute may be set.
        :param room_lookup: Lookup function to get the event's room from the data given, without this, a 'room' key will be ignored.
        :param speaker_lookup: Lookup function to get a conference user object for the data given, without this, speakers will be ignored.

        :returns: a new event or the existing one with fields updated from the given data
        :rtype: Event
        """
        assert isinstance(data, dict), 'Data must be a dictionary.'
        assert existing is None or isinstance(existing, cls), 'Existing instance must be an Event.'

        given_uuid = UUID(data.pop('guid')) if 'guid' in data else None
        if existing:
            obj = existing
            if assembly:
                assert obj.assembly == assembly, "Existing event's assembly does not match given one."
            if conference:
                assert obj.conference == conference, "Existing event's conference does not match given one."
            if given_uuid is not None:
                assert obj.pk == given_uuid, f"expected existing event's id {obj.pk} to match the given uuid {given_uuid}"
        else:
            obj = cls(assembly=assembly, conference=conference)
            if given_uuid is not None:
                obj.pk = given_uuid
        assert obj.conference_id is not None

        direct_fields = [
            'slug',
            'name',
            'abstract',
            'description',
            'language',
            'is_public',
            'schedule_start',
            'schedule_end',
        ]
        bool_fields = ['is_public']
        dt_fields = ['schedule_start', 'schedule_end']
        for fld in direct_fields:
            if fld in data:
                value = data[fld] if not pop_used_keys else data.pop(fld)
                if fld in bool_fields:
                    value = str2bool(value)
                if fld in dt_fields:
                    value = parse_datetime(value)
                    if not is_aware(value):
                        value = make_aware(value)
                setattr(obj, fld, value)
                if pop_used_keys:
                    del data[fld]

        if 'schedule_duration' in data:
            value = data['schedule_duration']
            obj.schedule_duration = str2timedelta(value) if not isinstance(value, timedelta) else value
            if pop_used_keys:
                del data['schedule_duration']

        if 'kind' in data:
            if allow_kind:
                obj.kind = data['kind']
            if pop_used_keys:
                del data['kind']

        if 'track' in data:
            if allow_track:
                track_q = Q(slug__iexact=data['track']) | Q(name__iexact=data['track'])
                trk = assembly.conference.tracks.get(track_q)
                obj.track = trk
            if pop_used_keys:
                del data['track']

        if 'room' in data:
            if room_lookup is not None:
                room = room_lookup(data['room'])
                obj.room = room
            else:
                raise RuntimeWarning('Event.from_dict() got data with "room" but no room_lookup was provided.')
            if pop_used_keys:
                del data['room']

        if 'speakers' in data:
            if speaker_lookup is not None:
                obj.save()  # TODO: check if we can do this better, but we need the model saved for adding speakers
                for order, item in enumerate(data['speakers']):
                    speaker = speaker_lookup(item)
                    if speaker is not None:
                        obj.ensure_speaker(speaker, order)
            else:
                raise RuntimeWarning('Event.from_dict() got data with "speakers" but no speaker_lookup was provided.')
            if pop_used_keys:
                del data['speakers']

        if pop_used_keys:
            obj.additional_data = data
        elif 'additional_data' in data:
            obj.additional_data = data['additional_data']

        if obj.additional_data and obj.additional_data.get('do_not_record', False) is True:
            obj.recording = Event.Recording.NO

        # load banner_image, if a URL is given (and has not changed since last time)
        if banner_image_url := data.get('banner_image_url'):
            if obj.banner_image_url != banner_image_url:
                try:
                    obj.load_banner_image_from_url(banner_image_url)
                except Exception:
                    # just log the error, if the event has no banner_image (yet), we don't care too much
                    logger.exception('Failed to download banner_image for (new) event %s: %s', obj.pk, banner_image_url)

        obj.clean()
        return obj

    def clean(self, test_for_publication: bool = False):
        super().clean()
        publication_errors = {}

        errors = {}

        sos = self.kind == self.Kind.SELF_ORGANIZED
        event_type = _('SoS') if sos else _('event')

        if Event.objects.exclude(id=self.id).filter(conference=self.conference, assembly=self.assembly, name=self.name).exists():
            errors['name'] = ValidationError(_('Event__name__already_exists %(event_type)s') % {'event_type': event_type}, code='duplicate')

        # SoS assembly setting on conference must be set
        if self.assembly and self.conference and sos:
            if not self.conference.self_organized_sessions_assembly_id:
                errors[NON_FIELD_ERRORS] = ValidationError(_('SoS__missing_assembly'), code='configuration')
                logging.error('Missing SoS assembly for %s!', self.conference.slug)
            else:
                self.assembly = self.conference.self_organized_sessions_assembly

        # ensure track is of the same conference
        if self.track_id is not None and self.track.conference_id != self.conference_id:
            errors['track'] = ValidationError(_('Event__track__needs_same_conference'), code='invalid')

        # ensure assembly is in the same conference
        if self.assembly and self.assembly.conference_id != self.conference_id:
            errors['assembly'] = ValidationError(_('Event__assembly__needs_same_conference'), code='invalid')

        # ensure room belongs to ...
        if self.room_id is not None:
            # ... the same conference
            if self.room.conference_id != self.conference_id:
                errors['room'] = errors.get('room', []).append(ValidationError(_('Event__room__needs_same_conference'), code='invalid'))

        # official events only in an official assembly
        if self.kind == Event.Kind.OFFICIAL and not self.assembly.is_official:
            errors['kind'] = ValidationError(_('Event__kind__official_assembly_only'), code='official_only')

        # When setting to public we must have a start
        if (test_for_publication or self.is_public) and not self.schedule_start:
            publication_errors['schedule_start'] = errors['schedule_start'] = ValidationError(
                _('Event__schedule_start__non_empty %(event_type)s') % {'event_type': event_type}, code='empty'
            )

        # start must be within conference timeframe
        if self.schedule_start is not None:
            if (self.conference.start is not None and self.conference.start > self.schedule_start) or (
                self.conference.end is not None and self.conference.end < self.schedule_start
            ):
                errors['schedule_start'] = ValidationError(
                    _('Event__schedule_start__outside_conference_time_frame %(event_type)s') % {'event_type': event_type}, code='invalid_date'
                )

        # When setting to public we must have a duration
        if (test_for_publication or self.is_public) and not self.schedule_duration:
            publication_errors['schedule_duration'] = errors['schedule_duration'] = ValidationError(
                _('Event__schedule_duration__non_empty %(event_type)s') % {'event_type': event_type}, code='empty'
            )
        if self.schedule_duration is not None:
            # duration must not be zero
            if self.schedule_duration <= timedelta(0):
                errors['schedule_duration'] = ValidationError(
                    _('Event__schedule_duration__must_be_greater_zero %(event_type)s') % {'event_type': event_type}, code='invalid'
                )

            # end cannot be after the end of the conference
            elif (
                self.conference is not None
                and self.conference.end is not None
                and self.schedule_start is not None
                and (self.schedule_start + self.schedule_duration) > self.conference.end
            ):
                errors['schedule_duration'] = ValidationError(
                    _('Event__schedule_end__after_conference_end %(event_type)s') % {'event_type': event_type}, code='invalid_schedule'
                )

        # Start and duration must be set if we want to go public
        if self.is_public and (not self.schedule_duration or not self.schedule_start):
            errors['is_public'] = ValidationError(
                _('Event__is_public__no_start_or_duration %(event_type)s') % {'event_type': event_type}, code='missing_values'
            )

        # Only allow events to be public if the conference has a start
        if (test_for_publication or self.is_public) and not self.conference.start:
            publication_errors[NON_FIELD_ERRORS] = errors[NON_FIELD_ERRORS] = ValidationError(
                _('Event__is_public__conference_has_no_start %(event_type)s') % {'event_type': event_type}, code='invalid_conference'
            )

        # if a room has been set, do some more checks
        if self.room is not None:
            # ensure schedule_duration etc. are set, if possible
            self._autoset_schedule_values()

            # search all candidate rooms
            possible_rooms = (
                Room.objects.assignable_in_timeframe(
                    conference=self.conference,
                    start=self.schedule_start,
                    end=self.schedule_end,
                    assembly=self.assembly,
                    is_sos=self.is_sos,
                )
                if self.schedule_start and self.schedule_end and self.schedule_duration and self.schedule_duration >= timedelta(seconds=1)
                else Room.objects.none()
            )

            # check that selected room is available
            if not possible_rooms.filter(id=self.room_id).exists():
                publication_errors['room'] = errors['room'] = ValidationError(_('Event__room__unavailable'), code='unavailable_room')

            else:
                # check for colliding events in the room
                conflicting_events = (
                    self.room.events.exclude(pk=self.pk)  # we don't conflict with ourselves
                    .exclude(schedule_end__lte=self.schedule_start)  # ignore events ending before our's
                    .exclude(schedule_start__gte=self.schedule_start + self.schedule_duration)  # ignore events starting after our's
                    .filter(Q(is_public=True) | Q(kind=Event.Kind.OFFICIAL))  # we only collide with public or official events (the latter might be non-public)
                )
                conflicts = conflicting_events.values('is_public', 'slug')

                if conflicts:
                    # construct error message listing the affected slugs and, if present, the number of reservations (non-public events w/ kind=official)
                    conflict_slugs = [x['slug'] for x in conflicts if x['is_public']]
                    conflict_slugs_str = ', '.join(f'"{slug}"' for slug in conflict_slugs) if conflict_slugs else ''
                    conflict_msg = _('Event__room__colliding_events_%(slugs)s') % {'slugs': conflict_slugs_str}
                    conflict_hidden = sum(1 for x in conflicts if not x['is_public'])
                    if conflict_hidden > 0:
                        if conflict_slugs:
                            conflict_msg += ' + '
                        conflict_msg += _('Event__room__colliding_events_plus%(count)dhidden') % {'count': conflict_hidden}

                    # signal the validation conflict
                    publication_errors['room'] = errors['room'] = ValidationError(
                        conflict_msg,
                        code='collision',
                        params={'slugs': conflict_slugs, 'hidden': conflict_hidden},
                    )

        if test_for_publication:
            raise ValidationError(publication_errors)
        if errors:
            raise ValidationError(errors)

    def save(self, *args, update_fields=None, **kwargs):
        self._autoset_schedule_values()

        if not self.slug:
            self.slug = gen_slug = slugify(self.name)[:50]
            while Event.objects.filter(conference=self.conference, slug=self.slug).exclude(pk=self.pk).exists():
                self.slug = gen_slug[:45] + '_' + ''.join([random.SystemRandom().choice('abcdefghijklmnopqrstuvwxyz0123456789') for i in range(4)])

        if update_fields is None or 'description' in update_fields:
            render_results = compile_translated_markdown_fields(self, self.conference, 'description')
            store_relationships(self.conference, self, render_results)

        return super().save(*args, update_fields=update_fields, **kwargs)

    def _autoset_schedule_values(self):
        if self.schedule_start is not None and self.schedule_duration is not None:
            self.schedule_end = self.schedule_start + self.schedule_duration
        elif self.schedule_start is not None and self.schedule_end is not None:
            self.schedule_duration = self.schedule_end - self.schedule_start
        else:
            self.schedule_end = None

    @property
    def conference_slug(self):
        return self.conference.slug

    @property
    def supports_sos_join(self):
        # we support joining only for SOS
        if self.kind != self.Kind.SELF_ORGANIZED:
            return False

        # we require a configured BigBlueButton integration
        return settings.INTEGRATIONS_BBB

    def ensure_speaker(self, user: PlatformUser, order: int):
        entry, created = self.participants.get_or_create(participant=user)
        if created:
            logger.info(
                'Added participant "%s" (%s) to event "%s" (%s), intended to be a speaker.',
                user.username,
                user.id,
                self.name,
                self.pk,
            )

        entry.role = EventParticipant.Role.SPEAKER
        entry.is_public = True
        entry.is_accepted = True
        entry.order = order
        entry.save()
        logger.info(
            'Forced participant #%s (%s) of event %s to accepted+public+speaker.',
            entry.id,
            user.username,
            self.pk,
        )

    def load_banner_image_from_url(self, url: str, save: bool = True):
        logger.debug('Loading banner_image for event %s (%s) from URL: %s', self.slug, self.pk, url)

        filename, data = download_from_url(url)
        target_filename = get_event_banner_filename(self, filename)

        logger.debug('Storing banner_image for user %s (%s): %s', self.slug, self.pk, target_filename)
        self.banner_image.save(target_filename, data)
        self.banner_image_url = url
        if save:
            self.save(update_fields=['banner_image', 'banner_image_url'])
        logger.debug('Updated banner_image for user %s (%s) as "%s" from URL: %s', self.slug, self.pk, target_filename, url)


def _EventAttachment_upload_path(instance, filename):
    # file will be uploaded to MEDIA_ROOT/user_<id>/<filename>
    return f'eventattachment/{instance.event.id}/{instance.id}'


class EventAttachment(models.Model):
    class Meta:
        verbose_name = _('EventAttachment')
        verbose_name_plural = _('EventAttachments')

    class Visibility(models.TextChoices):
        PRIVATE = 'private', _('EventAttachment__visibility-private')
        PARTICIPANTS = 'participants', _('EventAttachment__visibility-participants')
        CONFERENCE = 'conference', _('EventAttachment__visibility-conference')
        PUBLIC = 'public', _('EventAttachment__visibility-public')

    event = models.ForeignKey(Event, on_delete=models.CASCADE)
    id = models.UUIDField(default=uuid4, primary_key=True, editable=False)

    file = models.FileField(upload_to=_EventAttachment_upload_path)
    filename = models.CharField(max_length=255)
    mime_type = models.CharField(max_length=100)

    visibility = models.CharField(max_length=20, choices=Visibility.choices, default=Visibility.PRIVATE)


class EventParticipant(models.Model):
    class Meta:
        ordering = ['order']

    class Role(models.TextChoices):
        SPEAKER = 'speaker', _('EventParticipant__type-speaker')
        ANGEL = 'angel', _('EventParticipant__type-angel')
        REGULAR = 'regular', _('EventParticipant__type-regular')
        PROSPECT = 'prospect', _('EventParticipant__type-prospect')

    id = models.UUIDField(default=uuid4, primary_key=True, editable=False)

    event = models.ForeignKey(Event, on_delete=models.CASCADE, related_name='participants')
    participant = models.ForeignKey(PlatformUser, on_delete=models.CASCADE, related_name='events')

    role = models.CharField(
        max_length=20, choices=Role.choices, default=Role.REGULAR, help_text=_('EventParticipant__role__help'), verbose_name=_('EventParticipant__role')
    )

    # Future TODO: evaluate https://pypi.org/project/django-sortedm2m/ but for now we just use a simple order field
    order = models.PositiveIntegerField(default=0)

    is_accepted = models.BooleanField(default=False, help_text=_('EventParticipant__is_accepted__help'), verbose_name=_('EventParticipant__is_accepted'))
    is_public = models.BooleanField(default=False, help_text=_('EventParticipant__is_public__help'), verbose_name=_('EventParticipant__is_public'))

    public_description = models.TextField(
        blank=True, null=True, help_text=_('EventParticipant__public_description__help'), verbose_name=_('EventParticipant__public_description')
    )
    personal_comment = models.TextField(
        blank=True, null=True, help_text=_('EventParticipant__personal_comment__help'), verbose_name=_('EventParticipant__personal_comment')
    )

    def clean(self, *args, **kwargs):
        # verify that the participant is a member of the event's conference
        try:
            ConferenceMember.objects.get(conference=self.event.conference, user=self.participant)
        except ConferenceMember.DoesNotExist:
            raise ValidationError(_('EventParticipant__must_be_conference_member'))

        if not self.order:  # if 'order' is not explicitly set
            last_order = EventParticipant.objects.filter(event=self.event).aggregate(models.Max('order')).get('order__max')
            self.order = (last_order or 0) + 1

    def __str__(self):
        result = self.participant.username
        if self.role in [self.Role.SPEAKER, self.Role.ANGEL]:
            result += ' ' + self.get_role_display()
        return result
