import json
from collections import OrderedDict
from pathlib import Path
from urllib.parse import urlparse

import requests
from lxml import etree
from markdownify import markdownify
from requests_file import FileAdapter

from .base import filter_additional_data, schedule_time_to_timedelta
from .schedulejson import ScheduleJSON, ScheduleJSONSupport

s = requests.Session()
s.mount('file://', FileAdapter())


class ScheduleJSONFrabSupport(ScheduleJSONSupport):
    identifier = 'schedule-json-frab'
    readonly = True

    # fmt: off
    configuration_fields = {
        'feedback':        ('bool',   True, False, 'Enable/Disable feedback url generation'),
        'kind':            ('string', 'assembly', False, 'kind of events, either "assembly" or "official" or "sos" or "lightning"'),
        'images_base_url': (['bool', 'str'], False, False, 'Use image urls from origin instead of hub'),
        'host':            ('string', 'https://frab.cccv.de/', False, 'overwrite Base URL of the schedule, used for feedback page, images, etc.'),
        'auth':            ('string',  None, False, 'login data for frab instance, format: user:password'),
    }
    # fmt: on

    origin_base_url = None

    def ready(self):
        r = s.head(self.remote_url)
        return r.ok

    def fetch(self):
        """
        This method is the workhorse of the schedule support module:
        its job is to query upstream for the current set of data.

        It shall return a dictionary with keys 'rooms' and 'events',
        each containing a dictionary with entries mapping a source id
        to a dictionary which can be understood by Room.from_dict()
        and Event.from_dict() respectively.

        The hub will update events it already knows by the source id:
        all events need to have an unique but stable identifier, i.e.
        if the name of the event changes the identifier must not change!
        """

        instance = urlparse(self.remote_url)
        host = self.conf_value('host') or f'//{instance.hostname}'
        # e.g. https://frab.cccv.de/36c3/public/
        self.origin_base_url = f'https://{instance.hostname}{Path(instance.path).parent}/'
        frab = s

        if self.conf_value('auth'):
            page = etree.HTML(frab.get(host).text)
            user, password = self.conf_value('auth').split(':')
            login_data = {
                'user[email]': user,
                'user[password]': password,
                'user[remember_me]': 1,
                'authenticity_token': page.xpath("//meta[@name='csrf-token']")[0].get('content'),
            }
            frab.post(f'{host}/users/sign_in', login_data)

        schedule = ScheduleJSON.from_url(self.remote_url, frab)
        speakers = SpeakerJSON.from_url(self.remote_url.replace('schedule.json', 'speakers.json'), frab, base_url=host)

        events = {}
        kind = self.conf_value('kind')
        feedback = self.conf_value('feedback')

        # when images_base_url is True, use the base_url from the schedule.json body
        # otherwhise images_base_url is a string like https://frab.cccv.de/ or https://fahrplan.events.ccc.de/congress/2023/
        images_base_url = self.conf_value('images_base_url')
        if images_base_url is True:
            images_base_url = schedule.get('base_url')

        def overwrite_avatar_url(p):
            if p.get('avatar'):
                return {**p, 'avatar': p['avatar'].replace('/original/', '/large/').replace(f'{host}/', images_base_url)}
            return p

        for e in schedule.events():
            persons = [speakers.person(p.get('guid'), p) for p in e.get('persons', [])]

            computed_data = {'persons': persons}
            if feedback:
                computed_data['feedback_url'] = f"{self.origin_base_url}events/{e.get('id')}/feedback/new"

            # if images_base_url is set, we have to rewrite the public image URLs in additonal_data with the public CDN domain
            if images_base_url:
                computed_data['persons'] = [overwrite_avatar_url(p) for p in persons]
                if e.get('logo'):
                    computed_data['logo'] = f"{images_base_url}{e['logo'][1:]}"
            additional_data = filter_additional_data(e, computed_data)

            abstract = e.get('abstract') or ''
            description = e.get('description') or ''
            events[e.get('id')] = {
                'guid': e.get('guid'),
                'slug': e.get('slug').split(f"{e.get('id')}-")[1][0:150].strip('-') or e.get('slug')[0:150].strip('-'),
                'name': e.get('title'),
                'language': e.get('language'),
                'abstract': markdownify(abstract) if abstract else '',
                # TODO should we fetch translated descriptions from source?
                'description': markdownify(description) if description else '',
                'track': e.get('track'),
                'room': e.get('room'),
                'schedule_start': e.get('date'),
                'schedule_duration': str(schedule_time_to_timedelta(e.get('duration'))),
                'is_public': True,
                'kind': kind,
                'speakers': persons,
                # for the internal image urls when can use the origin host
                'banner_image_url': f"{self.origin_base_url}{e['logo'][1:]}" if e.get('logo') else None,
                'additional_data': additional_data,
            }

        return {
            'version': schedule.version(),
            'rooms': {r['name']: r for r in schedule.rooms()},
            'events': events,
        }

    def computed_data(self, event: dict):
        return {}


class SpeakerJSON:
    """
    Schedule Speakers from JSON document
    """

    _speakers = {}

    @classmethod
    def build_biogaphy(cls, s):
        res = f"{s.get('abstract') or ''} {s.get('description') or ''}".strip()
        if res:
            return markdownify(res)
        return ''

    def __init__(self, json, base_url):
        self._speakers = {
            s['guid']: OrderedDict(
                {
                    'guid': s['guid'],
                    'id': s['id'],  # TODO can we drop the old local integer ID? Ask native clients.  # noqa: ERA001
                    # we can't drop the id this time because we want to be able to redirect requests like https://fahrplan.events.ccc.de/congress/2023/fahrplan/speakers/19378.html # noqa: E501
                    'name': s.get('name'),
                    'public_name': s.get('public_name'),
                    'biography': SpeakerJSON.build_biogaphy(s),
                    'avatar': f"{base_url}{s['image']}" if s.get('image') else None,
                    'links': s.get('links', []),
                }
            )
            for s in json
        }

    @classmethod
    def from_url(cls, url, frab, base_url, language=None):
        r = frab.get(url, headers={'Accept-Language': language} if language else None)

        if r.ok is False:
            raise Exception(f'Request failed, HTTP {r.status_code}.')

        # maintain order from input file
        data = json.JSONDecoder(object_pairs_hook=OrderedDict).decode(r.text)

        # Close the raw file handle if it's still open
        if hasattr(r, 'raw') and r.raw.closed is False:
            r.raw.close()

        return SpeakerJSON(json=data.get('schedule_speakers').get('speakers'), base_url=base_url)

    def __getitem__(self, key):
        return self._speakers[key]

    def persons_map(self):
        return self._speakers

    def persons(self):
        return self._speakers.values()

    def person(self, guid, fallback=None):
        if guid in self._speakers:
            return self._speakers[guid]
        return fallback

    def __str__(self):
        return json.dumps(self._speakers.values(), indent=2)
